# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Choose k items from a set."""

import dataclasses
from typing import Any, List, Tuple, Union

from cascades._src.distributions import base
import jax
import jax.numpy as jnp


def _sample_log_probs(rng, log_probs, k=None):
  # Sample categorical distribution of log probs
  noise = jax.random.gumbel(rng, shape=log_probs.shape)
  perturbed = log_probs + noise
  if k is None:
    return jnp.argmax(perturbed, axis=-1)
  else:
    return jnp.argsort(perturbed, axis=-1)[:k]


@dataclasses.dataclass(frozen=True, eq=True)
class Choose(base.Distribution):
  """Choose k of n options. Uses gumbel top-k trick."""
  k: int = 1
  options: Union[Tuple[Any], List[Any]] = tuple()  # pytype: disable=annotation-type-mismatch

  def sample(self, rng) -> base.RandomSample:
    rng = base.get_rng(rng)
    n = len(self.options)

    log_p = jnp.log(1.0 / n)
    log_probs = jnp.full((n,), fill_value=log_p)

    chosen_idxs = _sample_log_probs(rng, log_probs=log_probs, k=self.k)
    chosen = []
    for idx in chosen_idxs:
      chosen.append(self.options[idx])
    return_value = base.RandomSample(value=chosen, log_p=log_p * self.k)
    return return_value



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sampling from OpenAI api."""
import bisect
import dataclasses
import functools
import os
from typing import Iterable, Optional, Text
import uuid

from cascades._src.distributions import base as dists
import openai

openai.api_key = os.getenv('OPENAI_API_KEY')


# TODO(ddohan): Persist cache to disk
@functools.lru_cache()
def cached_completion(rng=None, **kwargs):
  del rng
  return openai.Completion.create(**kwargs)


@dataclasses.dataclass(eq=True, frozen=True)
class GPT(dists.Distribution):
  """Sample a String from GPT."""

  prompt: Text = ''

  stop: Optional[Iterable[Text]] = ('\n',)

  engine = 'davinci-codex'
  temperature: float = 0.7
  max_tokens: int = 128
  top_p: float = .95
  frequency_penalty: int = 0
  presence_penalty: int = 0

  def sample(self, rng=None, raw=False):
    """Sample a value from the distribution given rng key.

    Args:
      rng: Optional random key.
      raw: If True, return OpenAI api request verbatim.

    Returns:
      a RandomSample or, if raw is True, the OpenAI response dict.
    """
    if rng is None:
      rng = uuid.uuid4().hex
    elif not isinstance(rng, int):
      raise ValueError('RNG must be an integer or Jax key.')
    result = cached_completion(
        rng=rng,
        model=self.engine,
        prompt=self.prompt,
        temperature=self.temperature,
        max_tokens=self.max_tokens,
        top_p=self.top_p,
        frequency_penalty=self.frequency_penalty,
        presence_penalty=self.presence_penalty,
        logprobs=0,
        stop=self.stop,
        echo=False,
    )['choices'][0]

    completion = result['text']
    logprobs = result['logprobs']

    span = (0, len(completion) + 1)
    start = 0
    end = bisect.bisect_left(logprobs['text_offset'], span[1])
    total_logprob = sum(result['logprobs']['token_logprobs'][start:end])
    if raw:
      return (total_logprob, result)
    return dists.RandomSample(score=total_logprob, value=completion)

  def log_prob(self, value, raw=False):
    """Get log prob of completion.

    Args:
      value: Completion to score.
      raw: If True, return raw OpenAI API response.

    Returns:
      float logprob or the OpenAI response dict.
    """
    text = self.prompt + value
    span = (len(self.prompt), len(text))
    result = cached_completion(
        rng=0,
        model=self.engine,
        prompt=text,
        temperature=self.temperature,
        max_tokens=0,
        echo=True,
        top_p=self.top_p,
        frequency_penalty=self.frequency_penalty,
        presence_penalty=self.presence_penalty,
        logprobs=0,
    )['choices'][0]
    logprobs = result['logprobs']
    start = bisect.bisect(logprobs['text_offset'], span[0]) - 1
    end = bisect.bisect_left(logprobs['text_offset'], span[1])
    excerpt = text[span[0]:span[1]]
    joined = ''.join(logprobs['tokens'][start:end])
    assert excerpt in joined
    total_logprob = sum(logprobs['token_logprobs'][start:end])
    if raw:
      return (total_logprob, result)
    return total_logprob



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Effect handlers which specify how to interpret yielded effects."""
import collections
import dataclasses
import math
import types
from typing import Any, Callable, Dict, List, Optional, Text, Type, Union

from cascades._src.distributions import base as dists
import immutabledict
import jax

Distribution = dists.Distribution
RandomSample = dists.RandomSample

## Methods to run a generator.


def run_with_intermediates(gen):
  """Run generator to completion and return final value and intermediates.

  Args:
    gen: Generator.

  Returns:
    Dict with value = return_value and intermediates = yielded values.
  """
  assert isinstance(gen, types.GeneratorType)
  intermediates = []
  while True:
    try:
      x = next(gen)
      intermediates.append(x)
    except StopIteration as e:
      return_value = None
      if e.args is not None and len(e.args):
        return_value = e.args[0]
      break
  return dict(return_value=return_value, intermediates=intermediates)


def forward_sample(fn, seed, *args, **kwargs):
  """Sample once from fn(*args, **kwargs) with seed.

  Args:
    fn: Callable which creates a generator, yielding distributions or effects.
    seed: Random seed to use.
    *args: Args passed to fn.
    **kwargs: Kwargs passed to fn.

  Returns:
    Tuple of (handlers, result), where handlers is the effect handler stack
    used to interpret the model, and result is a dict containing `result` and
    `intermediates`.
  """
  # TODO(ddohan): add ApplyContext back in.
  # gen_fn = ApplyContext(lambda: fn(*args, **kwargs))
  gen_fn = lambda: fn(*args, **kwargs)
  # TODO(ddohan): Add in ParamHandler
  handler_fn = compose_handlers([Record, StopOnReject, Observer, Sampler])
  forward_sample_handler: Record = handler_fn(Seed(seed=seed, gen_fn=gen_fn))  # pytype: disable=annotation-type-mismatch
  result_with_metadata = forward_sample_handler.run_with_intermediates(
      verbose=False)
  result_with_metadata['observed_likelihood'] = (
      forward_sample_handler.observed_likelihood)
  return forward_sample_handler, result_with_metadata


def rejection_sample(fn, seed, max_attempts, *args, **kwargs):
  """Sample repeatedly until a success."""
  rng = dists.get_rng(seed)
  for i in range(max_attempts):
    rng, subrng = jax.random.split(rng)
    handlers, result = forward_sample(fn, subrng, *args, **kwargs)
    observed_likelihood = handlers.observed_likelihood
    if not math.isinf(observed_likelihood):
      result['attempts'] = i + 1
      return handlers, result
  return None


## Helper methods to create effects within a model.
# TODO(ddohan): Move these to a separate file.


def log(value, name=None):
  """Record a value into a trace."""
  yield Log(value=value, name=name or 'log')


def sample(dist=None, obs=None, name=None):
  """Sample value from distribution. Optionally takes observed value."""
  effect = yield from sample_and_score(dist=dist, obs=obs, name=name)
  return effect.value


def observe(dist=None, obs=None, name=None):
  """Observe that given distribution takes on observed value. Returns score."""
  effect = yield from sample_and_score(dist=dist, obs=obs, name=name)
  return effect.score


def sample_and_score(dist=None, obs=None, name=None):
  """Get RandomSample from distribution. Optionally takes an observed value."""
  if obs is None:
    effect = Sample(fn=dist, value=None, name=name)
  else:
    effect = Observe(fn=dist, value=obs, name=name)

  returned_effect = yield effect
  return returned_effect


def factor(score, name=None):
  """Add score to likelihood of current trace.

  Args:
    score: Numeric value to add to trace likelihood. Used to intervene in trace
      weights.
    name: Name for the effect.

  Yields:
    An Observe effect for the given Factor distribution.
  """
  dist = dists.Factor(factor=score)
  yield from sample(dist=dist, obs=dists.FactorSentinel(), name=name)


def reject(reason, name=None):
  """Add -inf term to likelihood of current trace."""
  dist = dists.Factor(reason=reason, factor=-math.inf)
  yield Reject(
      fn=dist,
      value=dists.FactorSentinel,
      score=-math.inf,
      name=name or 'reject')


def param(name=None, dist=None, value=None):
  """Create a parameter. Samples from `dist` by default if value not given."""
  effect = yield Param(fn=dist, value=value, name=name)
  return effect.value


def _yielded_value_to_effect(value):
  """Convert a value to an effect."""
  if isinstance(value, Effect):
    return value
  elif isinstance(value, (Distribution, dists.np_dists.Distribution)):
    # pylint does not handle dataclass inheritance properly.
    return Sample(fn=value)  # pylint: disable=unexpected-keyword-arg
  else:
    raise ValueError('Unknown effect type %s' % str(value))


## Basic Effect types.


@dataclasses.dataclass
class Effect:
  """Track state of an effect which is processed by EffectHandlers."""
  # Unique name for site which yielded this effect.
  name: Optional[Text] = None

  # Results of sampling or scoring.
  value: Optional[Any] = None
  score: Optional[float] = None

  # Callable, generally used for sampling and scoring.
  fn: Optional[Union[Callable, Distribution]] = None  # pylint: disable=g-bare-generic
  args: Optional[List[Any]] = dataclasses.field(default_factory=list)
  kwargs: Optional[Dict[Text, Any]] = dataclasses.field(default_factory=dict)

  # If True, then the generator should be halted
  # generally used when likelihood becomes infinite.
  should_stop: bool = False
  replayed: bool = False

  metadata: Optional[Any] = None


@dataclasses.dataclass
class Log(Effect):
  """Log information to the trace."""
  value: Any = None

  # Unique name for site which yielded this effect.
  name: Optional[Text] = None


@dataclasses.dataclass
class Sample(Effect):
  """A distribution together with its sampled value."""
  pass


@dataclasses.dataclass
class Observe(Effect):
  """A distribution together with its observed value."""
  pass


@dataclasses.dataclass
class Reject(Observe):
  """Reject a trace. Equivalent to -inf likelihood."""
  pass


@dataclasses.dataclass
class Param(Observe):
  """Model parameters."""
  pass


## Core effect handlers.


class EffectHandler:
  """Wraps generator fn and customizes how yielded effects are interpreted."""

  def __init__(self, gen_fn=None):
    """Wraps given generator fn. May be None if self.__call__ is not used."""
    self.gen_fn = gen_fn
    self._context = None

  def run_with_intermediates(self, fn_or_gen=None, verbose=None):
    return run_with_intermediates(self(fn_or_gen=fn_or_gen, verbose=verbose))

  def process(self, effect):
    """Applied before yield."""
    return effect

  def postprocess(self, effect):
    """Applied after yield."""
    return effect

  def on_return(self, return_value):
    """Runs on return from program."""
    pass

  def get_stack(self):
    """If wrapped gen_fn is an EffectHandler, get entire stack of handlers."""
    handlers = []
    handler = self
    # TODO(ddohan): Unwrap functools.partial as well.
    while isinstance(handler, EffectHandler):
      handlers.append(handler)
      handler = handler.gen_fn
    return handlers

  def __call__(self, fn_or_gen=None, verbose=False, nested=False):
    """Iterate generator fn, interpreting effects using (post)process.

    Args:
      fn_or_gen: If given, generator or callable returning a generator to trace.
        If not given, defaults to self.gen_fn
      verbose: If True, log out the trace.
      nested: Used internally for recursive nested calls. Changes behavior of

    Yields:
      Effects

     # Indent to satisfy pylint.
     Returns:
       Value returned from wrapped generator.
    """

    def _log(x):
      if verbose:
        print(x)

    if fn_or_gen is None:
      fn_or_gen = self.gen_fn
    if fn_or_gen is None:
      raise ValueError(
          '`gen_fn` must be passed as argument EffectHandler __init__ or __call__. Was None.'
      )

    if isinstance(fn_or_gen, types.GeneratorType):
      # assert not args and not kwargs, (args, kwargs)
      gen = fn_or_gen
    else:
      # Call function to get a generator back
      gen = fn_or_gen()

    value_to_inject = None
    return_value = None
    try:
      while True:
        _log(f'Injecting {value_to_inject}')

        yielded_value = gen.send(value_to_inject)

        _log(f'Yielded value: {yielded_value}')
        if isinstance(yielded_value, types.GeneratorType):

          # Recursively trace through yielded generator.
          # TODO(ddohan): Support yielding functions as well.
          value_to_inject = yield from self(
              fn_or_gen=yielded_value, nested=nested + 1)
          _log(f'received from yield: {value_to_inject}')
          continue
        else:
          effect = _yielded_value_to_effect(yielded_value)

          # Process & postproces modify the effect in place
          effect = self.process(effect)
          if effect is None:
            raise ValueError(f'Did not return effect from {self}')
          yield effect

          # TODO(ddohan): postprocess will be applied from outside in
          # Do we want to somehow apply it inside out as well?
          # This would require an additional yield.
          effect = self.postprocess(effect)
          if effect is None:
            raise ValueError(f'Did not return effect from {self}')
          if effect.should_stop:
            return None
          value_to_inject = effect
    except StopIteration as e:
      if e.args is not None and len(e.args):
        return_value = e.args[0]
        _log(e.args)
        return return_value
      return None
    finally:
      # Always ensure that any locally scoped state is unloaded.
      # TODO(ddohan): Test that this behaves as expected for WithContext
      # handlers when `nested=True`.
      if not nested:
        self.on_return(return_value)


def compose_handlers(handlers: List[Type[EffectHandler]]):
  """Compose together a list of handlers."""
  if not handlers:
    raise ValueError('Cannot compose an empty set of handlers.')

  def init_handlers(gen_fn=None):
    for handler_cls in handlers[::-1]:
      if handler_cls is None:
        continue
      gen_fn = handler_cls(gen_fn=gen_fn)
    return gen_fn

  return init_handlers


class Record(EffectHandler):
  """Record effects into trace."""

  def __init__(self, gen_fn=None):
    super().__init__(gen_fn=gen_fn)
    self.trace = collections.OrderedDict()
    self.keys = []
    self.observed_likelihood = 0.0
    self.unobserved_likelihood = 0.0
    self.return_value = None
    self.done = False

  def __getitem__(self, key):
    """Get variable from trace by name."""
    return self.trace[key]

  @property
  def joint_likelihood(self):
    return self.observed_likelihood + self.unobserved_likelihood

  def __repr__(self):
    kvs = [f'  {k}: {v}' for k, v in self.trace.items()]
    kvs = '\n'.join(kvs)
    return f'Record(\n{kvs}\n)'

  def on_return(self, return_value):
    """Runs on return from program."""
    self.result = return_value
    self.done = True

  def process(self, effect):
    if effect.name is None:
      raise ValueError(
          f'Cannot record trace for effect without a name: {effect}')
    if effect.name in self.trace:
      raise ValueError(f'Address `{effect.name}` is already in trace')
    self.trace[effect.name] = effect
    self.keys.append(effect.name)
    if isinstance(effect, Observe):
      if effect.score is not None:
        self.observed_likelihood += effect.score
    elif isinstance(effect, Sample):
      self.unobserved_likelihood += effect.score
    return effect


class Seed(EffectHandler):
  """Set `rng` kwarg for each effect."""

  def __init__(self, gen_fn=None, seed=None):
    super().__init__(gen_fn=gen_fn)
    if isinstance(seed, int):
      self.rng = jax.random.PRNGKey(seed=seed)
    else:
      self.rng = seed

  def process(self, effect):
    self.rng, subrng = jax.random.split(self.rng)
    effect.kwargs['rng'] = subrng
    return effect


class Sampler(EffectHandler):
  """Sample values from Distributions."""

  def __init__(self, gen_fn=None, await_timeout=None):
    super().__init__(gen_fn=gen_fn)
    self._await_timeout = await_timeout

  def process(self, effect):
    if isinstance(effect, Sample):
      if effect.value is None:
        random_sample = dists.sample_distribution(
            effect.fn,
            *effect.args,
            await_timeout=self._await_timeout,
            **effect.kwargs)
        effect.value = random_sample.value
        effect.score = random_sample.log_p
    return effect


class Observer(EffectHandler):
  """Compute score of Observe statements."""

  def __init__(self, gen_fn=None, await_timeout=None, rescore=True):
    super().__init__(gen_fn=gen_fn)
    self._await_timeout = await_timeout
    self._rescore = rescore

  def process(self, effect):
    if isinstance(effect, Observe):
      if effect.value is None:
        raise ValueError(f'Observe with a None value: {effect}')
      if isinstance(self._rescore, bool):
        should_rescore = self._rescore
      else:
        should_rescore = effect.name in self._rescore

      if should_rescore:
        score = dists.score_distribution(
            effect.fn,
            effect.value,
            await_timeout=self._await_timeout,
        )
      else:
        score = None
      effect.score = score
    return effect


class StopOnReject(EffectHandler):
  """Mark that tracing should stop in event of inf likelihood."""

  def process(self, effect):
    if (isinstance(effect, Observe) and effect.score is not None):
      should_stop = jax.numpy.any(jax.numpy.isinf(effect.score))
      if jax.device_get(should_stop):
        effect.should_stop = True
    return effect


class AutoName(EffectHandler):
  """Assign a unique name to each effect based on index in trace."""

  def __init__(self, gen_fn=None):
    super().__init__(gen_fn=gen_fn)
    self.idx = 0

  def __repr__(self):
    return f'<AutoName {self.gen_fn}>'

  def process(self, effect):
    name = effect.name
    if name is None:
      name = effect.fn.__class__.__name__
    name = f'{self.idx}/{name}'
    effect.name = name
    self.idx += 1
    return effect


class Replay(EffectHandler):
  """Look up site values from given trace."""

  def __init__(self,
               gen_fn,
               trace,
               assert_all_used=True,
               replay_scores=False,
               rescore=False,
               await_timeout=None):
    super().__init__(gen_fn=gen_fn)
    if not isinstance(trace, immutabledict.ImmutableOrderedDict):
      trace = immutabledict.ImmutableOrderedDict(trace)
    self.trace = trace
    self._unused = set(k for k in trace)
    self._assert_all_used = assert_all_used
    if replay_scores and rescore:
      raise ValueError('Cannot both replay and rescore log probs.')
    self._replay_scores = replay_scores
    self._rescore = rescore
    self._await_timeout = await_timeout

  def on_return(self, return_value):
    """Runs on return from program."""
    if self._assert_all_used and self._unused:
      raise ValueError(f'Some keys were unused in replay: {self._unused}')

  def process(self, effect):
    if effect.name in self.trace:
      cached = self.trace[effect.name]
      effect.value = cached.value
      if self._replay_scores:
        effect.score = cached.score
      elif self._rescore:
        effect.score = dists.score_distribution(
            fn=effect.fn, value=effect.value, await_timeout=self._await_timeout)
      effect.replayed = True
      self._unused.remove(effect.name)
    return effect



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic distribution types and helpers."""
from concurrent import futures
import dataclasses
import functools
from typing import Any, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from numpyro import distributions as np_dists

DEFAULT_TIMEOUT = 60

## Base Distribution and RandomSample types.


@dataclasses.dataclass(eq=True, frozen=True)
class Distribution:
  """Each Distribution implements at least `sample` and `score`."""
  # stand in for kw_only in py3.10
  capture: Any = dataclasses.field(repr=False, default=None)

  name: Any = dataclasses.field(repr=True, default=None)
  observed: Any = dataclasses.field(repr=True, default=None)

  def __post_init__(self):
    if self.capture is not None:
      raise ValueError('Use kwargs for all arguments.')

  def sample(self, rng):
    """Draw a sample from the distribution."""
    raise NotImplementedError

  def prob(self, value):
    """Score a sample by probability under the distribution."""
    raise NotImplementedError

  def log_prob(self, value):
    """Score a sample by log probability under the distribution."""
    return jax.numpy.log(self.prob(value))

  def score(self, value):
    """Score sample. Default to `log_prob` but may be any unnormalized value."""
    return self.log_prob(value)

  def support(self):
    """Possible values which can be drawn from the distribution.

    Not easily defined for all distributions.
    """
    raise NotImplementedError


# TODO(ddohan): Can it both be frozen & do post_init casting?
# @dataclass(frozen=True, eq=True)
@dataclasses.dataclass(eq=True)
class RandomSample:
  """A value with an optional score (log_p) and source distribution."""
  # stand in for kw_only in py3.10
  capture: Any = dataclasses.field(repr=False, default=None)

  log_p: Optional[Union[float, jax.numpy.DeviceArray]] = None
  value: Any = None
  dist: Optional[Union[Distribution, functools.partial]] = None

  def __post_init__(self):
    if self.capture is not None:
      raise ValueError('Use kwargs for all arguments.')

  @property
  def score(self):
    """Potentially unnormalized score. Defaults to log_p."""
    return self.log_p


## Helper methods


# TODO(ddohan): Consider moving get_rng to separate file.
def get_rng(seed):
  """Get jax prng key from seed. Does nothing is seed is a PRNGKey."""
  if isinstance(seed, int):
    return jax.random.PRNGKey(seed)
  # TODO(ddohan): What's the right type for jax seed?
  # if isinstance(seed, chex.PRNGKey):

  # Seed is already a jax.random.PRNGKey, so we just pass it through.
  return seed


def score_distribution(fn: Union[Distribution, np_dists.Distribution],
                       value: Any,
                       await_timeout: Optional[int] = None):
  """Score value under distribution.

  Args:
    fn: Object which defines a `score` or `log_prob` method.
    value: Value to score.
    await_timeout: Length of time to wait. Defaults to DEFAULT_TIMEOUT.

  Returns:
    Float likelihood of value under distribution.
  """
  if hasattr(fn, 'score'):
    score_fn = fn.score
  elif hasattr(fn, 'log_prob'):
    score_fn = fn.log_prob
  else:
    raise ValueError('Must defined `score` or `log_prob` methods.')

  if not callable(score_fn):
    raise ValueError(f'Score method {score_fn} is not callable on {fn}')
  score = score_fn(value)

  if isinstance(score, futures.Future):
    score = score.result(timeout=await_timeout or DEFAULT_TIMEOUT)
  return score


def sample_distribution(fn, *args, await_timeout=None, **kwargs):
  """Sample value from function or distribution.

  If `fn` comes from third party distribution library (e.g tf distributions or
  NumPyro), then sample using library specific method. Otherwise, if fn is
  callable, draw a sample by calling fn.

  Args:
    fn: Callable or object which defines a `score` or `log_prob` method.
    *args: Args passed to sample fn.
    await_timeout: Length of time to wait. Defaults to DEFAULT_TIMEOUT.
    **kwargs: Kwargs passed to sample fn.

  Returns:
    Sampled value, with (potentially unnormalized) likelihood.
  """
  if isinstance(fn, np_dists.Distribution):
    # Numpyro distributions.
    key = kwargs['rng']
    kwargs['key'] = key
    del kwargs['rng']
    value = fn.sample(*args, **kwargs)
    log_p = fn.log_prob(value)
    return RandomSample(value=value, log_p=log_p)
  elif hasattr(fn, 'sample') and callable(fn.sample):
    fn = fn.sample
  random_sample = fn(*args, **kwargs)
  if isinstance(random_sample, futures.Future):
    random_sample = random_sample.result(
        timeout=await_timeout or DEFAULT_TIMEOUT)
  if not isinstance(random_sample, RandomSample):
    raise ValueError(
        f'Expected sample to return RandomSample. Got {random_sample}')
  return random_sample


## Factor distributions.


# TODO(ddohan): Consider switching sentinel to 0-dim array
class FactorSentinel:
  pass


@dataclasses.dataclass(eq=True, frozen=True)
class Factor(Distribution):
  """Add given `score` to observed likelihood.

  Used to manually add terms to model likelihood. Always returns the given
  `score` for `observe`.
  """
  factor: float = 0.0
  reason: Any = None

  def sample(self, rng):
    del rng
    return RandomSample(value=FactorSentinel(), log_p=self.factor)

  def score(self, value):
    del value
    return self.factor


## Discrete distributions


@dataclasses.dataclass(frozen=True, eq=True)
class UniformCategorical(Distribution):
  """Return a random choice from the distribution."""
  options: Tuple[Any] = tuple()  # pytype: disable=annotation-type-mismatch

  def sample(self, rng=None) -> RandomSample:
    idx = jax.random.randint(rng, (), 0, len(self.options))
    idx = int(idx)
    sample = self.options[idx]
    log_p = -jnp.log(len(self.options))
    return_value = RandomSample(value=sample, log_p=log_p)
    return return_value

  def log_prob(self, value):
    del value
    # TODO(ddohan): Assert that sample is one of the options
    log_p = -jnp.log(len(self.options))
    return log_p

  def support(self):
    return self.options



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sample values from a model."""
from functools import partial  # pylint: disable=g-importing-member
from typing import Any

from cascades._src import handlers as h
import jax

# Inference


# TODO(ddohan): Add goal conditioning (future prompting) back in.
def build_tracer(
    fn,
    seed: Any,  # h.RNGKey,
    trace=(),
    reparam_fn=None,
    rescore=False) -> h.Record:
  """Make a tracer which samples values and records to a tape.

  Args:
    fn: Function to trace.
    seed: Seed for clone of trace.
    trace: Trace of site name -> Effect to replay values.
    reparam_fn: Partial Reparam handler which can be used to reparameterize
      distributions.
    rescore: Whether to rescore the observed samples.

  Returns:
    New handler stack instantiated around the function.
  """
  # TODO(ddohan): Allow context handlers within program.
  # gen_fn = h.ApplyContext(fn)
  gen_fn = fn

  handler_stack = [
      # Record effects to a tape
      h.Record,
      # Sample/observe distributions & await futures
      h.StopOnReject,
      partial(h.Observer, rescore=rescore),
      h.Sampler,
      # Replace the known values when generator is rerun
      partial(h.Replay,
              # Thread RNG through effect.
              trace=trace,
              replay_scores=True,
              assert_all_used=True),
      partial(h.Seed, seed=seed),
  ]
  handler_fn = h.compose_handlers(handler_stack)
  if reparam_fn is None:
    reparam_fn = lambda x: x
  record_handler: h.Record = handler_fn(reparam_fn(gen_fn))
  return record_handler


class Sampler:
  """Run sampling on a model."""

  def __init__(
      self,
      model,
      # examples: Optional[Iterable[Dict[Text, Any]]] = None,
      # observe: Optional[Dict[Text, Any]] = None,
      # future_prompt=False,
      rescore=True,
      reparam_fn=None,
  ):
    self._model = model
    # self._examples = examples or ()
    # self._observed = observe or dict()
    # self._future_prompt = future_prompt
    self._rescore = rescore
    self._reparam_fn = reparam_fn

  def build_tracer(self, seed=0):
    """Instantiate a tracer for given seed. Does not evalute concrete values."""
    # TODO(ddohan): Handle observed variables, with and without future prompting
    tracer = build_tracer(
        self._model,
        seed=seed,
        # examples=self._examples,
        # observed=self._observed,
        # future_prompt=self._future_prompt,
        rescore=self._rescore,
        reparam_fn=self._reparam_fn)
    tracer.done = False
    return tracer

  def reify(self, seed=0):
    tracer = self.build_tracer(seed=seed)
    reify(tracer)
    return tracer

  def parallel(self, pool, seed, n):
    """Run function in parallel."""
    seeds = jax.random.split(jax.random.PRNGKey(seed), n)
    tracers = []
    for seed in seeds:
      tracer = self.build_tracer(seed=seed)
      f = pool.submit(reify, tracer=tracer)
      tracer.future = f
      tracers.append(tracer)
    return tracers


def reify(tracer: h.Record, verbose=False):
  """Run tracer and return the result. Makes implicit values explicit.

  Modifies `tracer` in place.

  Args:
    tracer: Tracer to run.
    verbose: If true, log out effects.

  Returns:
    Result of running the program.
  """
  it = tracer()
  return_value = None
  while True:
    try:
      eff = next(it)
      if verbose:
        print(eff)
    except StopIteration as e:
      if e.args is not None and len(e.args):
        return_value = e.args[0]
        if verbose:
          print(f'Return value: {eff}')
      break
  tracer.done = True
  tracer.return_value = return_value
  return return_value




# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cascades API."""

# A new PyPI release will be pushed everytime `__version__` is increased
# When changing this, also update the CHANGELOG.md
__version__ = '0.2.0'

from cascades._src.distributions.base import UniformCategorical
from cascades._src.distributions.choose import Choose
from cascades._src.distributions.strings import get_default_lm
from cascades._src.distributions.strings import mock_lm
from cascades._src.distributions.strings import String
from cascades._src.handlers import factor
from cascades._src.handlers import log
from cascades._src.handlers import observe
from cascades._src.handlers import param
from cascades._src.handlers import reject
from cascades._src.handlers import rejection_sample
from cascades._src.handlers import sample
from cascades._src.sampler import Sampler



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for sampler."""

from absl.testing import absltest
from cascades._src import handlers as h
from cascades._src import sampler
import jax.numpy as jnp
from numpyro import distributions as np_dists


def _binomial(k, p=0.5):
  total = 0
  for _ in range(k):
    flip = yield h.sample(np_dists.Bernoulli(probs=p))
    total += flip
  return total


def _flip_paths(n, a, b):
  nheads = yield _binomial(n)
  if int(nheads) != a:
    yield h.reject(reason=f'nheads {nheads} != {a}')

  nheads = yield _binomial(n)
  if int(nheads) != b:
    yield h.reject(reason=f'nheads {nheads} != {b}')


def _gaussian_mixture(locs):
  """Standard gaussian mixture model."""
  n = len(locs)
  mixing_dist = np_dists.Categorical(probs=jnp.ones(n) / n)
  component_dist = np_dists.Normal(loc=jnp.array(locs), scale=jnp.ones(n))
  mixture = np_dists.MixtureSameFamily(mixing_dist, component_dist)
  return mixture


def gaussian_mixture_likelihood(proposal_loc=0.0,
                                proposal_scale=3.0,
                                mixture_locs=(-5.0, 5.0)):
  """Demonstrate proposing & scoring in same program."""
  # Proposal distribution
  proposal = yield h.sample(
      name='proposal',
      dist=np_dists.Normal(loc=proposal_loc, scale=proposal_scale))

  mixture = _gaussian_mixture(mixture_locs)

  # Add term to likelihood
  yield h.sample(name='score', dist=mixture, obs=proposal)

  return proposal


class SamplerLikelihood(absltest.TestCase):

  def test_likelihood_weighting(self):
    """Sample from normal, and weight using mixture."""
    locs = [-5.0, 5.0]

    def _fn(verbose=False):
      del verbose
      return gaussian_mixture_likelihood(mixture_locs=locs)

    s = sampler.Sampler(
        model=lambda: h.AutoName(_fn)(), rescore=True, reparam_fn=None)  # pylint: disable=unnecessary-lambda
    tracer: h.Record = s.build_tracer(0)
    self.assertFalse(tracer.done)
    sampler.reify(tracer)
    mixture = _gaussian_mixture(locs)
    expected_score = mixture.log_prob(tracer.return_value)
    self.assertEqual(expected_score, tracer.observed_likelihood)
    self.assertTrue(tracer.done)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for string distributions."""

from absl.testing import absltest
from cascades._src import handlers as h
from cascades._src import sampler
from cascades._src.distributions import strings


class StringsTest(absltest.TestCase):

  def test_mock_lm(self):

    def fn():
      a = yield h.sample(
          strings.String(lm=strings.mock_lm('ABC'), until=None, k=2), name='a')
      b = yield h.sample(
          strings.String(lm=strings.mock_lm('XYZ'), until=None, k=1), name='b')
      return a + b

    s = sampler.Sampler(fn)
    trace = s.reify(seed=0)
    self.assertEqual('ABCABCXYZ', trace.return_value)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""String distributions."""

from concurrent import futures
import dataclasses
import math
from typing import Any, Text

from cascades._src.distributions import base as dists


DEFAULT_LM_CLIENT = dict(default=None)


def set_default_lm(lm):
  """Set the default language model client."""
  DEFAULT_LM_CLIENT['default'] = lm


def get_default_lm():
  """Return a default language model client."""
  lm = DEFAULT_LM_CLIENT.get('default')
  if not lm:
    raise ValueError('No default LM set.')
  return lm


def removeafter(text: Text, token: Text = '==='):
  """Truncate `text` after until `token`."""
  idx = text.find(token)
  if idx >= 0:
    text = text[:idx + len(token)]
    return text
  else:
    return None


def mock_lm(response: Text) ->...:
  """Return Distribution class which will return a fixed response."""

  @dataclasses.dataclass(frozen=True)
  class MockLM(dists.Distribution):
    prompt: Text = ''

    def sample(self, rng):
      del rng
      return dists.RandomSample(value=response, log_p=-1.0)

    def prob(self, value):
      return 1.0

  return MockLM


# TODO(ddohan): Use cascade machinery + yields here as well
def iterate_until(rng,
                  prompt: Text,
                  lm,
                  k: int = 2,
                  until: Text = '\n',
                  reject_if_missing=True,
                  timeout=60,
                  rescore=True):
  """Sample from language model up to `k` times.

  Stop at given `until` token.

  Args:
    rng: Random seed.
    prompt: Prefix text to condition on.
    lm: Language model distribution to use.
    k: Number of times to iterate the sampling.
    until: Sample until this token is found.
    reject_if_missing: If True, then reject the trace if `until` is not found.
    timeout: Maximum time to wait for a generation.
    rescore: If true, then rescores the generation after truncation, as an
      `Observe`

  Returns:
    None if rejected, otherwise the generated text.
  """
  full_generation = ''
  likelihood = 0.0
  for _ in range(k):
    current_prompt = prompt + full_generation
    step_lm = lm(prompt=current_prompt)
    generation = step_lm.sample(rng=rng)
    if isinstance(generation, futures.Future):
      generation = generation.result(timeout)
    full_generation += generation.value
    likelihood += generation.log_p
    if until:
      clipped = removeafter(text=full_generation, token=until)
      if clipped is not None:
        if rescore:
          # print(f"rescore: `{clipped}`")
          score = lm(prompt=prompt).score(clipped)
          if isinstance(score, futures.Future):
            score = score.result(timeout)
        else:
          score = 0.0
        # preclip = clipped
        clipped = clipped[:-len(until)]
        # print(f'`{preclip}` -> `{clipped}`')
        return dists.RandomSample(log_p=score, value=clipped)
  if until and reject_if_missing:
    # We were looking for something and we didn't find it.
    # Use a -inf likelihood
    likelihood = -math.inf
  return dists.RandomSample(log_p=likelihood, value=full_generation)


@dataclasses.dataclass(eq=True, frozen=True)
class String(dists.Distribution):
  """Sample a String from an LM. May iterate calls to the LM multiple times."""

  prompt: Text = ''
  k: int = 2
  until: Text = '\n'
  lm: Any = None
  timeout: int = 60

  def _get_lm(self):
    return self.lm or get_default_lm()

  def sample(self, rng):
    rng = dists.get_rng(rng)
    value = iterate_until(
        rng=rng,
        lm=self._get_lm(),
        prompt=self.prompt,
        until=self.until,
        k=self.k,
        timeout=self.timeout)
    return value

  def log_prob(self, value):
    if not isinstance(value, str):
      value = f' {value}'
    if self.until:
      value = value + self.until
    return self._get_lm()(prompt=self.prompt).score(sample=value)
